toJSON <- function (x, scalars=TRUE)  {
  n <- names(x)
  if (is.factor(x) || inherits(x, "POSIXt")) {
    if (is.null(n)) x <- as.character(x) else {
      x <- as.character(x)
      names(x) <- n
    }
  }
  if (is.null(x))
    return("null")
  if (!is.null(n)) { # named vectors
    if (anyDuplicated(n))
      stop("A JSON list must have unique names")
    n <- quote.string(n)
    return (paste("{", paste('"',n,'":',unlist(lapply(x,toJSON)), collapse=',', sep=''), "}", sep=''))
  }
  ## convert arrays to recursive lists so a[x,y,z] becomes a[x][y][z] in JS
  if (is.array(x)) return (paste("[",paste(apply(x, 1, toJSON),collapse=','),"]",sep=''))
  if (is.numeric(x)) {
    fin <- is.finite(x)
    if (all(fin)) { # fast track - no NA/NaNs,...
      if (scalars && length(x) == 1) return(as.character(x))
      return (paste('[', paste(x, collapse=','), ']', sep=''))
    }
    x <- as.character(x)
    x[!fin] <- paste('"',x[!fin],'"',sep='')
    if (scalars && length(x) == 1) return(as.character(x))
    return (paste('[', paste(x, collapse=','), ']', sep=''))
  }
  if (is.logical(x)) {
    x <- c("false","true")[as.integer(x) + 1L]
    if (any(is.na(x))) x[is.na(x)] <- '"NA"'
    if (scalars && length(x) == 1) return(x)
    return (paste('[', paste(x, collapse=','), ']', sep=''))
  }
  x <- quote.string(x)
  if (any(is.na(x))) x[is.na(x)] <- '"NA"'
  if (scalars && length(x) == 1) return(paste('"',x,'"',sep=''))
  paste('[', paste('"', x, '"', sep='', collapse=','), ']', sep='')
}

