/* quote string vectors as required by JSON specs. */

#define USE_RINTERNALS 1
#include <Rinternals.h>
#include <stdlib.h>
#include <string.h>

static char sbuf[65536];

static const char hex[16] = "0123456789ABCDEF";

static SEXP do_quote(const char *str, int nq, int len) {
    char *buf = sbuf, *c;
    const char *end = str + len;
    SEXP res;

    if (nq * 5 + len > sizeof(sbuf)) {
	buf = malloc(nq * 5 + len);
	if (!buf) Rf_error("out of memory, cannot allocate %d bytes for a string buffer", nq * 5 + len);
    }
    c = buf;
    while (str < end) {
	if (*str == '"' || *str == '\\') { /* quote " and \ */
	    *(c++) = '\\';
	    *(c++) = *(str++);
	} else if (*str > 31) {/* most common case - no translation */
	    *(c++) = *(str++);
	} else { /* 0..31 to be quoted */
	    *(c++) = '\\';
	    switch (*str) {
	    case '\t': *(c++) = 't'; break;
	    case '\n': *(c++) = 'n'; break;
	    case '\r': *(c++) = 'r'; break;
	    case '\b': *(c++) = 'b'; break;
	    default:
		*(c++) = 'u';
		*(c++) = '0';
		*(c++) = '0';
		*(c++) = hex[(*str) >> 4];
		*(c++) = hex[(*str) & 15];
	    }
	    str++;
	}
    }
    /* FIXME: if we fail here out malloc will leak */
    res = Rf_mkCharLenCE(buf, (int)(c - buf), CE_UTF8);
    if (buf != sbuf) free(buf);
    return res;
}

SEXP quote_string(SEXP str) {
    SEXP res = 0;
    int i = 0, n = LENGTH(str);
    while (i < n) {
	SEXP se = STRING_ELT(str, i);
#ifdef IS_UTF8 /* slight shortcut for R 2.15.0 and higher */
	const char *c = IS_UTF8(se) ? CHAR(se) : Rf_translateCharUTF8(se), *s = c;
#else
	const char *c = Rf_translateCharUTF8(se), *s = c;
#endif
	int nq = 0;
	while (*c) {
	    if ((*c > 0 && *c < 31) || *c == '\\' || *c == '"') nq++;
	    c++;
	}
	if (nq) { /* number of needed quotes */
	    if (!res) {
		res = PROTECT(Rf_allocVector(STRSXP, n));
		memcpy(STRING_PTR(res), STRING_PTR(str), sizeof(SEXP) * n);
	    }
	    SET_STRING_ELT(res, i, do_quote(s, nq, (int) (c - s)));
	}
	i++;
    }
    if (!res) return str;
    UNPROTECT(1);
    return res;
}

SEXP quote_raw(SEXP str) {
    SEXP res = 0;
    int i = 0, n = LENGTH(str);
    const char *c = (const char*) RAW(str), *s = c;
    int nq = 0;
    while (i < n) {
	/* FIXME: we should also quote 0x80 .. 0xff since there is no attached encoding */
	if ((c[i] >= 0 && c[i] < 31) || c[i] == '\\' || c[i] == '"') nq++;
	i++;
    }
    res = PROTECT(Rf_allocVector(STRSXP, 1));
    SET_STRING_ELT(res, 0, nq ? do_quote(s, nq, (int) (c - s)) : mkChar(s));
    UNPROTECT(1);
    return res;
}
